/**
 * Node.java
 * @author John Green
 * Aug 27, 2004
 * www.joanju.com
 *
 * Copyright (C) 2004 Joanju Limited. All rights reserved.
 */
package com.joanju;

import com.joanju.ProparseLdr;


/** Sample class for loading Proparse's syntax tree into a Java structure.
 */
public class Node {

	public Node(int handle) {
		this.nodeHandle = handle;
		this.type = parser.getNodeTypeI(nodeHandle);
	}

	static private ProparseLdr parser = ProparseLdr.getInstance();

	private int nodeHandle = 0;
	private int type = 0;
	private Node down;
	private Node parent;
	private Node right;


	public static Node getTree() {
		int topHandle = parser.getHandle();
		parser.nodeTop(topHandle);
		return getTree(topHandle, null);
	}

	private static Node getTree(int inHandle, Node parent) {
		Node thisNode = new Node(inHandle);
		thisNode.parent = parent;
		int handle = parser.getHandle();
		if (parser.nodeFirstChildI(inHandle, handle) != 0) {
			thisNode.down = getTree(handle, thisNode);
			handle = parser.getHandle();
		}
		if (parser.nodeNextSiblingI(inHandle, handle) != 0) {
			thisNode.right = getTree(handle, parent);
		} else {
			parser.releaseHandle(handle);
		}
		return thisNode;
	}



	//// Member funcs



	/** Get an attribute from proparse */
	public int attrGet(int key) {
		return parser.attrGetI(nodeHandle, key);
	}



	/** Set an attribute in proparse */
	public void attrSet(int key, int val) {
		parser.attrSet(nodeHandle, key, val);
	}



	public Node firstChild() { return down; }



	/** Find the first direct child with a given node type. */
	public Node findDirectChild(int nodeType) {
		for (Node node = firstChild(); node!=null; node = node.nextSibling()) {
			if (node.getType() == nodeType) return node;
		}
		return null;
	}



	/**
	 * First Natural Child is found by repeating firstChild() until a natural node is found.
	 * If the start node is a natural node, then it is returned.
	 * The test for "natural" node is line number. Synthetic node line numbers == 0.
	 * Note: This is very different than Prolint's "NextNaturalNode" in lintsuper.p.
	 */
	public Node firstNaturalChild() {
		if (getLine()>0) return this;
		for (Node n = down; n!=null; n = n.down) {
			if (n.getLine()>0) return n;
		}
		return null;
	}



	public int getColumn() {
		return parser.getNodeColumn(nodeHandle);
	}



	public int getFileIndex() {
		return parser.getNodeFileIndex(nodeHandle);
	}



	public String getFilename() {
		return parser.getNodeFilename(nodeHandle);
	}



	/** Get the integer "handle" to the node within Proparse.dll */
	public int getHandle() { return nodeHandle; }



	/** Get the line number within the source file where this node was found.
	 * Will return zero for "synthetic" nodes.
	 */
	public int getLine() {
		return parser.getNodeLine(nodeHandle);
	}



	/** Return int[3] of nodes file/line/col. */
	public int[] getPos() {
		return new int[] {getFileIndex(), getLine(), getColumn()};
	}



	/** Get the token text for this node */
	public String getText() {
		return parser.getNodeText(nodeHandle);
	}


	
	/** Get the integer node type */
	public int getType() { return type; }

	
	
	/** Get the node's type name (the text representation of the integer node type) */
	public String getTypeName() { return parser.getTokenTypeName(type); }


	
	/** Return the last immediate child (no grandchildren). */
	public Node lastChild() {
		Node ret = firstChild();
		if (ret==null) return null;
		while (ret.nextSibling()!=null) ret = ret.nextSibling();
		return ret;
	}



	public Node lastDescendant() {
		Node ret = lastChild();
		for (Node temp = ret; temp!=null; temp = ret.lastChild()) {
			ret = temp;
		}
		return ret;
	}


	
	public Node nextSibling() { return right; }



	public Node parent() { return parent; }



}
