/**
 * TreeView.java
 * @author John Green
 * Aug 27, 2004
 * www.joanju.com
 *
 * Copyright (C) 2004 Joanju Limited. All rights reserved.
 */
package com.joanju;

import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.FileDialog;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.MessageBox;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.TreeItem;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.SashForm;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Tree;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.widgets.Text;
import org.eclipse.swt.widgets.Label;


/** A simple dialog which displays Proparse's parse tree
 * for the input filename.
 */
public class TreeView {

	private final Display display = Display.getDefault();
	private ProparseLdr parser = ProparseLdr.getInstance();
	private String windowTitle = "Proparse Tree Browser";
	private Text nodeColumn;
	private Text nodeLine;
	private Text nodeText;
	private Text typeText;
	private Text typeInt;
	private Text nodeFile;
	private Tree tree;
	protected Shell shell;

	
	
	/** Parse a file and display the tree.
	 * Filename is optional, will be prompted for if not input as an argument.
	 * Loads schema from file "proparse.schema" in the working directory, if present.
	 * (See Proparse documentation, "Proparse without Progress".)
	 * Loads configuration by "proparse.config", if present.
	 * @see com.joanju.ProparseUtil#readConfig(ProparseLdr, String)
	 */
	public static void main(String[] args) {
		TreeView window = new TreeView();
		window.parser.configSet("PROPATH", "."); // In case there's no config file.
		try {
			ProparseUtil.readConfig(window.parser, "proparse.config");
		} catch (Exception e) {}
		window.parser.schemaOldLoad();
		window.open(args.length==0 ? null : args[0]);
	}

	
	
	protected void displayNode(org.eclipse.swt.widgets.Widget widget) {
		if (! (widget instanceof TreeItem)) return;
		Node node = (Node) ((TreeItem)widget).getData();
		typeInt.setText(Integer.toString(node.getType()));
		typeText.setText(node.getTypeName());
		nodeText.setText(node.getText());
		nodeFile.setText(node.getFilename());
		nodeLine.setText(Integer.toString(node.getLine()));
		nodeColumn.setText(Integer.toString(node.getColumn()));
	}

	
	
	private boolean fillTree(String filename) {
		if (parser.parse(filename) < 0) {
			showMessage(parser.errorGetText());
			return false;
		}
		Node topNode = Node.getTree();
		TreeItem item = new TreeItem(tree, 0);
		item.setText(topNode.getTypeName()); // Will always be "Program_root"
		item.setData(topNode);
		fillTreeSub(topNode.firstChild(), item);
		return true;
	}
	
	
	
	private void fillTreeSub(Node currNode, TreeItem parentTreeItem) {
		TreeItem currItem = new TreeItem(parentTreeItem, 0);
		currItem.setText(currNode.getTypeName() + " " + currNode.getText());
		currItem.setData(currNode);
		Node next;
		if ((next=currNode.firstChild()) != null) fillTreeSub(next, currItem);
		if ((next=currNode.nextSibling()) != null) fillTreeSub(next, parentTreeItem);
	}



	/** Parse and display the contents for the input filename.
	 * Input null to have the file dialog displayed to prompt for filename.
	 * This method assumes that Proparse.dll configuration settings and schema
	 * have already been done.
	 */
	public void open(String filename) {
		createContents();
		String useName = filename;
		if (useName==null) {
			FileDialog fileDialog = new FileDialog(shell);
			fileDialog.open();
			useName = fileDialog.getFilterPath() + "/" + fileDialog.getFileName();
			if (useName.equals("")) return;
		}
		if (!fillTree(useName)) {
			shell.dispose();
			return;
		}
		tree.addListener (SWT.Selection, new Listener () {
			public void handleEvent (Event event) {
				displayNode(event.item);
			}
			});
		shell.open();
		while (!shell.isDisposed()) {
			if (!display.readAndDispatch())
				display.sleep();
		}
	}

	

	private void showMessage(String theMessage) {
		MessageBox dialog = new MessageBox(shell, SWT.ICON_INFORMATION);
		dialog.setMessage(theMessage);
		dialog.setText(windowTitle);
		dialog.open();
	} // showMessage()



	protected void createContents() {
		shell = new Shell();
		shell.setLayout(new FillLayout(SWT.HORIZONTAL));
		shell.setText(windowTitle);
		shell.setImage(
			new Image(
				display
				, getClass().getClassLoader().getResourceAsStream("frog32.gif") ) );

		final SashForm sashForm = new SashForm(shell, SWT.HORIZONTAL);

		final Composite left = new Composite(sashForm, SWT.NONE);
		left.setLayout(new FillLayout());

		tree = new Tree(left, SWT.BORDER);

		final Composite right = new Composite(sashForm, SWT.NONE);
		final GridLayout gridLayout = new GridLayout(2, false);
		right.setLayout(gridLayout);

		final Label label = new Label(right, SWT.RIGHT);
		final GridData gridData_2 = new GridData(GridData.HORIZONTAL_ALIGN_FILL);
		label.setLayoutData(gridData_2);
		label.setText("Int Type");

		typeInt = new Text(right, SWT.BORDER);
		final GridData gridData = new GridData(GridData.HORIZONTAL_ALIGN_FILL | GridData.FILL_HORIZONTAL);
		typeInt.setLayoutData(gridData);

		final Label label_1 = new Label(right, SWT.RIGHT);
		label_1.setLayoutData(new GridData(GridData.HORIZONTAL_ALIGN_FILL));
		label_1.setText("Type");

		typeText = new Text(right, SWT.BORDER);
		final GridData gridData_1_1 = new GridData(GridData.HORIZONTAL_ALIGN_FILL);
		typeText.setLayoutData(gridData_1_1);

		final Label label_1_1 = new Label(right, SWT.RIGHT);
		label_1_1.setLayoutData(new GridData(GridData.HORIZONTAL_ALIGN_FILL));
		label_1_1.setText("Text");

		nodeText = new Text(right, SWT.BORDER);
		final GridData gridData_1_2 = new GridData(GridData.HORIZONTAL_ALIGN_FILL);
		nodeText.setLayoutData(gridData_1_2);

		final Label label_1_2 = new Label(right, SWT.RIGHT);
		label_1_2.setLayoutData(new GridData(GridData.HORIZONTAL_ALIGN_FILL));
		label_1_2.setText("File");

		nodeFile = new Text(right, SWT.BORDER);
		final GridData gridData_1 = new GridData(GridData.HORIZONTAL_ALIGN_FILL);
		nodeFile.setLayoutData(gridData_1);

		final Label label_column = new Label(right, SWT.RIGHT);
		label_column.setLayoutData(new GridData(GridData.HORIZONTAL_ALIGN_FILL));
		label_column.setText("Line");

		nodeLine = new Text(right, SWT.BORDER);
		final GridData gridData_1_3 = new GridData(GridData.HORIZONTAL_ALIGN_FILL);
		nodeLine.setLayoutData(gridData_1_3);

		final Label label_1_2_2 = new Label(right, SWT.RIGHT);
		label_1_2_2.setLayoutData(new GridData(GridData.HORIZONTAL_ALIGN_FILL));
		label_1_2_2.setText("Column");

		nodeColumn = new Text(right, SWT.BORDER);
		final GridData gridData_1_4 = new GridData(GridData.HORIZONTAL_ALIGN_FILL);
		nodeColumn.setLayoutData(gridData_1_4);
		sashForm.setWeights(new int[] { 1, 1 });
	} // createContents

	
	
}
